/**/

   /* =============================================================================================== */
   /*  rpminfo.cmd                                                                                    */
   /*  Anton Monroe <amonroe5301@gmail.com>                                                           */
   /*  August 2024                                                                                    */
   /*                                                                                                 */
   /*  This calls rpm.exe to look up information about an rpm archive or rpm-installed file           */
   /*  I wrote it because I was baffled by the rpm command line options, most of which                */
   /*  I did not need to know about. To get information from rpm.exe, you must pass it a              */
   /*  very long and rather complicated query format string. This program lets me                     */
   /*  define what I want by simply filling in some blanks in a Rexx procedure, then it               */
   /*  generates the format strings and calls rpm.exe.                                                */
   /*                                                                                                 */
   /*  The alternative is to use rpm's popt aliases, which means writing the format                   */
   /*  strings by hand. See /usr/lib/rpm/rpmpopt* for some examples.                                  */
   /*                                                                                                 */
   /* =============================================================================================== */

   /*  The comments like "fold00" are for the benefit of the folding text editor I use.               */
   /*  If you use eFTE you should see my folds instead of the comments.                               */

   /* ==== NOTES ==================================================================================== */ /*fold00*/
   /*                                                                                                 */
   /*  an rpm package contains a lot of information in the form of tags. For instance,                */
   /*  querying the FILES tag returns a list of files in the package. There are a few                 */
   /*  hundred possible tags, and they can be displayed any way you want using an                     */
   /*  elaborate system of format strings. You can do anything you want, =if= you                     */
   /*  can figure out the right rpm.exe command line to do it.                                        */
   /*                                                                                                 */
   /*  RPMINFO.CMD gives me a convenient way to extract the information I am likely to want           */
   /*  in a readable format, without needing to remember how to use rpm.exe. Most of what it          */
   /*  does is putting together a very long and incomprehensible format string that                   */
   /*  it passes to rpm.exe.                                                                          */
   /*                                                                                                 */
   /*  rpm actually provides its own way of doing that with 'popt' aliases. I tried                   */
   /*  adding items to my /etc/popt file, but gave up on that idea.                                   */
   /*                                                                                                 */
   /*  How it works:                                                                                  */
   /*  the command line switches tell rpminfo.cmd to ask rpm about one or more things.                */
   /*  The results from running rpm.exe are saved in an array, and when all the queries               */
   /*  are done the array is displayed.                                                               */
   /*                                                                                                 */
   /*  Most of the procedures have names that start with "Add" or "Qf".                               */
   /*  A Qf* procedure returns a format string for displaying one or more tags. For instance,         */
   /*  QfChangelog() returns a format string for a set of tags related to the changelog,              */
   /*  like CHANGELOGNAME and CHANGELOGTEXT.                                                          */
   /*                                                                                                 */
   /*  An Add* procedure represents a query, usually about some group of tags.                        */
   /*  ('Add' because it adds lines to the output array). If you run                                  */
   /*      rpminfo.cmd <filename> -sets="package changelog"                                           */
   /*  this is what happens:                                                                          */
   /*      the Main level calls AddSets() which                                                       */
   /*         (1) calls QfPackage() which                                                             */
   /*            builds format strings for tags NAME, SUMMARY, URL, and others.                       */
   /*            then returns a long format string for that set of tags.                              */
   /*         (2) calls QfChangelog() which                                                           */
   /*            builds format strings for tags CHANGELOGNAME, CHANGELOGTEXT, etc.                    */
   /*            then returns a format string for that set of tags.                                   */
   /*         (3) runs rpm.exe, and saves the results to the array.                                   */
   /*         (4) may also insert or change lines in the array                                        */
   /*         (5) calls PrintArray() which                                                            */
   /*             sends the whole array to the screen                                                 */
   /*                                                                                                 */
   /*  I have tried to make it easy to add or change things. Each Add* procedure is                   */
   /*  independent of the others. AddSets() knows about several defined sets of tags,                 */
   /*  but you create your own set by writing a new Qf* procedure. Or you can write a                 */
   /*  new Add* procedure.                                                                            */
   /*                                                                                                 */
   /*  That should be enough to let you edit rpm format strings. Try changing some tag                */
   /*  or Set variables and see what happens. Keep reading if you want to know more.                  */
   /*                                                                                                 */
   /*                                                                                                 */
   /*    The Qf* routines are all based on a template, so you can add features by copying             */
   /*    one of them and editing it. The Set template represents a set of tags. A set has             */
   /*    at least one tag in it.                                                                      */
   /*                                                                                                 */
   /*    The 'tag.' stem variable represents one tag, where                                           */
   /*          tag.ID              is the tag name. For example '{NAME}'                              */
   /*          tag.width           if not '' specifies a fixed width for tag.ID. Example '-10'        */
   /*          tag.prefix          text to insert before tag.ID                                       */
   /*          tag.suffix          text to insert after tag.ID                                        */
   /*          tag.qf              is the complete format string for this tag                         */
   /*                                                                                                 */
   /*    You define the tag by editing the first four variables. For instance,                        */
   /*            tag.suffix = '\n'                                                                    */
   /*    adds a linefeed at the end of the tag format string.                                         */
   /*                                                                                                 */
   /*    You can use Rexx variables and functions to define elements                                  */
   /*            tag.ID = '{'translate(foo)'}'                                                        */
   /*                                                                                                 */
   /*    Then the format string for that tag is built by the line in the template that says           */
   /*         "tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix"                              */
   /*    Note the double '%%'; because I am using 4os2 which will eat one '%'. See below.             */
   /*                                                                                                 */
   /*    A set of tags can be grouped inside a Set template.  The 'Set.' stem variable                */
   /*    has only two parts, the square brackets that start and end the expansion part                */
   /*    of the string. See below about expansion                                                     */
   /*                                                                                                 */
   /*          Set.prefix = '['                                                                       */
   /*          Set.suffix = ']'                                                                       */
   /*                                                                                                 */
   /*    Keep the square brackets, but you can add things to the Set prefix or suffix--               */
   /*            Set.suffix = Set.suffix||'\n'                                                        */
   /*    adds a linefeed at the end of the entire Set format string                                   */
   /*                                                                                                 */
   /*    So the complete queryformat string for a set with one tag is built by adding the             */
   /*    Set prefix and suffix to the tag format string--                                             */
   /*            Set.qf = Set.prefix||tag.qf||Set.suffix                                              */
   /*                                                                                                 */
   /*    Here is the empty template procedure "QfTemplate()". Notice that I reuse                     */
   /*    the 'tag.' stem, defining each tag and adding it to 'Set.qf' as I go, then                   */
   /*    add the prefix and suffix last.  That makes adding more tags to a set easy and               */
   /*    practically foolproof-- just copy and paste an empty tag block                               */
   /*                                                                                                 */
   /*    QfTemplate: procedure                                                                        */
   /*    Set. = ''                                                                                    */
   /*    tag. = ''                                                                                    */
   /*                                                                                                 */
   /*    Set.prefix = '['                                                                             */
   /*    Set.suffix = ']'                                                                             */
   /*                                                                                                 */
   /*  TAG1                                                                                           */
   /*    tag.ID = '{}'                                                                                */
   /*    tag.width = ''                                                                               */
   /*    tag.prefix = ''                                                                              */
   /*    tag.suffix = ''                                                                              */
   /*    tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                                     */
   /*    Set.qf = Set.qf||tag.qf                                                                      */
   /*                                                                                                 */
   /*  TAG2                                                                                           */
   /*    tag.ID = '{}'                                                                                */
   /*    tag.width = ''                                                                               */
   /*    tag.prefix = ''                                                                              */
   /*    tag.suffix = ''                                                                              */
   /*    tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                                     */
   /*    Set.qf = Set.qf||tag.qf                                                                      */
   /*                                                                                                 */
   /*  TAG3                                                                                           */
   /*    tag.ID = '{}'                                                                                */
   /*    tag.width = ''                                                                               */
   /*    tag.prefix = ''                                                                              */
   /*    tag.suffix = ''                                                                              */
   /*    tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                                     */
   /*    Set.qf = Set.qr||tag.qf                                                                      */
   /*                                                                                                 */
   /*    Set.qf = Set.prefix||Set.qf||Set.suffix                                                      */
   /*    return Set.qf                                                                                */
   /* ----------------------------------------------------------------------------------------------- */

   /*  If you want to make any adjustments to the output of rpm.exe, the PostProcessor                */
   /*  routine is probably the place to do it.                                                        */

   /*                                                                                                 */
   /*  fix: allow separate output for each rpmPkgN, with default of 'stdout'. that is                 */
   /*              "rpminfo.cmd *.rpm -sets=requires -outfile=*.requires"                             */
   /*       would send separate output for 'abc.rpm' to 'abc.rpm.requires'                            */
   /*       in the meantime, I must do it from the command line--                                     */
   /*              for fileN in (*.rpm) (set outfile=%@filename[%fileN].requires %+ ...               */
   /*                                                                                                 */

   /***************************************************************************************************/
   /*  akm:                                                                                           */
   /*  These are some notes I made as I figured things out.  Mostly from the                          */
   /*  Suse 7.3 Linux docs (in /usr/share/doc/packages), which were much easier                       */
   /*  to understand than what rpm provided                                                           */
   /*                                                                                                 */
   /*  An rpm package has 'tags' that contain various kinds of information.                           */
   /*  A tag is an array of items.                                                                    */
   /*  The contents of a tag are printed according to a queryformat string.                           */
   /*  The string is very much like a printf format string. A tag is specified as                     */
   /*          %{TAG}  where 'TAG' is the tag name.                                                   */
   /*                                                                                                 */
   /*  The '%' makes things tricky with an OS/2 shell like cmd.exe or 4os2.exe                        */
   /*  For 4os2 you can escape the '%' by typing %%{TAG} or by backquoting                            */
   /*  the whole string. Or better yet, do all the string processing in Rexx.                         */
   /*                                                                                                 */
   /*  Examples of format strings:                                                                    */
   /*          %{FILENAMES}                                                                           */
   /*                is the first item of the FILENAMES array.                                        */
   /*                                                                                                 */
   /*  To expand an array, put it in square brackets:                                                 */
   /*          [%{FILENAMES}]                                                                         */
   /*                expands to all items of the array, with nothing to separate the items.           */
   /*                That is not very readable.                                                       */
   /*          [%{FILENAMES}\n]                                                                       */
   /*                prints a linefeed after each item                                                */
   /*          abc\n[%{FILENAMES}\n]                                                                  */
   /*                prints "abc" on one line, then each filename on a line                           */
   /*          [abc %{FILENAMES}\n]                                                                   */
   /*                for each filename, prints "abc <filename>"                                       */
   /*          [%{FILEMTIMES:date}        %{FILENAMES}\n]                                             */
   /*                for each file, prints the file's date+time and the filename                      */
   /*                                                                                                 */
   /*  Some tag arrays have only one item.                                                            */
   /*          [%{=NAME}        %{FILENAMES}\n]                                                       */
   /*                for each file in the package, prints the package name and filename.              */
   /*                The NAME array only has one item; the '=' tells RPM to keep printing             */
   /*                the first NAME for each FILENAME                                                 */
   /*                                                                                                 */
   /*  Some tags can take modifiers. You can show the file times in different ways with               */
   /*          [%{FILEMTIMES}        %{FILENAMES}\n]                                                  */
   /*  or                                                                                             */
   /*          %{FILEMTIMES:date} changes how the FILEMTIMES is displayed                             */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*                                                                                                 */
   /*    (copied from Suse Linux docs)                                                                */
   /*          Simple conditionals may be evaluated through query expressions. Expressions            */
   /*          are delimited by %|...|. The only type of expression currently supported               */
   /*          is a C-like ternary conditional, which provides simple if/then/else                    */
   /*          conditions. For example, the following query format display "present" if               */
   /*          the SOMETAG tag is present, and "missing" otherwise:                                   */
   /*                                                                                                 */
   /*              %|SOMETAG?{present}:{missing}|                                                     */
   /*                                                                                                 */
   /*          Notice that the subformats "present" and "missing" must be inside of curly             */
   /*          braces.                                                                                */
   /*          Example: (from popt configuration)                                                     */
   /*                  %|PACKAGER?{Packager    : %{PACKAGER}\n}|\                                     */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*    akm: looking at /usr/lib/rpm/rpmpopt* might give some clues                                  */
   /*    it shows the queryformats that correspond to rpm options, like:                              */
   /*            rpm   alias --changelog --qf '\                                                      */
   /*                    [* %{CHANGELOGTIME:day} %{CHANGELOGNAME}\n%{CHANGELOGTEXT}\n\n]' \           */
   /*                    --POPTdesc=$"list change logs for this package"                              */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*                                                                                                 */
   /*   to get a list of packages:                                                                    */
   /*        rpm.exe -q -a                                                                            */
   /*   to get a file list of a package called gcc:                                                   */
   /*        rpm.exe -q --list gcc                                                                    */
   /*   list with more info:                                                                          */
   /*        rpm.exe -q --dump gcc                                                                    */
   /*   to list all files with package basename:                                                      */
   /*        rpm.exe -q -a --filesbypkg                                                               */
   /*   to list all tags rpm can use in queryformat, with no description                              */
   /*        rpm.exe --querytags                                                                      */
   /*   I have not found a complete or convenient list of what the various tags mean                  */
   /*                                                                                                 */
   /***************************************************************************************************/

   Main: /*fold00*/
   /* =============================================================================================== */
   /*  this block has some variables you might want to change:                                        */
   /*    default list of what sets to show                                                            */
   DefaultSetList = 'Package Files Changelog Requires RequiresByFile Recommends' ,
                    'Provides Obsoletes Scripts'
   /*                                                                                                 */
   /*  where to send output and error/progress messages                                               */
   /*  Logfile can be changed on the command line, which is only useful for debugging                 */
   Outfile = 'stdout'
   Logfile = 'stderr'
   /*                                                                                                 */
   /*   some strings that are used to divide sections of the Rexx output array                        */
   Str. = ''
   Str.0divFile = copies('=', 80)
   Str.0divSet = copies('-', 80)
   Str.0divTag = copies('.', 80)
   /*                                                                                                 */
   /*  with '\n' for use in format strings                                                            */
   Str.0SetSuffix = Str.0divSet||'\n'
   /* =============================================================================================== */

   /* Load RexxUtil Library                                                                           */
   if RxFuncQuery('SysLoadFuncs') then do
         call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
         call SysLoadFuncs
      end
   signal on NoValue
   /* =============================================================================================== */
   /*  the plan:                                                                                      */
   /*  rpminfo.cmd works by writing an array of text (the stem variable called 'Output')              */
   /*  Most of the text is produced by calling rpm.exe and capturing the output                       */
   /*  rpm.exe must be called at least one time for each file, and maybe more.                        */
   /*  rpminfo.cmd can modify the text further, then send it to the screen or file(s).                */
   /*                                                                                                 */
   /*  this loop starts writing the array, then calls various 'Addxxx' subroutines                    */
   /*  which add lines to it. Each Addxxx is responsible for building an rpm query                    */
   /*  statement, calling rpm.exe, and doing whatever post-processing it wants.                       */
   /*  When the work is done, call PrintArray to send the array to screen or file.                    */
   /* =============================================================================================== */

   parse arg Mainargs
   call Init

   if Opt.0debug > 0 then call LogIt 'There are 'Flist.0 'matching files'

   do fileCtr = 1 to Flist.0
      rpmPkgN = translate(Flist.fileCtr, "/", "\")
      rpmPkgNShort = filespec('name', rpmPkgN)

      if FileExt(rpmPkgN) == 'rpm' then
         TargetSwitch = '--package'
      else
         TargetSwitch = '--file'
      if Opt.0progress > 0 then call ShowProgress 'reading 'rpmPkgNShort

      /*    add header with name of rpm file                                                          */
      call Out Str.0divFile
      if DirExist(Flist.fileCtr) then
         call Out 'directory: 'rpmPkgN
      else
         call Out 'file: 'rpmPkgN
      call Out Str.0divFile

      if Opt.0taglist \= '' then do
            /*    get entire list of tag names in this package                                        */
            call AddTagList
         end
      if Opt.0tags \= '' then do
            /*  get tags the user asked for                                                           */
            call AddUserTags
         end
      /*  get sets of tags                                                                            */
      call AddSets
      if Opt.0progress > 0 then call ShowProgress ''
      if Opt.0debug > 0 then call LogIt 'there are 'OutText.0 'lines in array'
      call Out ''
      /*  show the results for this file, reinitialize output array                                   */
      if Opt.0test > 0 then do
            call LogIt '[Test] Done' Flist.fileCtr
         end
      else do
            call PrintArray 'OutText.'
         end
      OutText. = ''
      OutText.0 = 0
   end fileCtr
   call Egress 'okay'

   /* === End of Main =============================================================================== */

   Init: /*fold00*/
   /* =============================================================================================== */
   /*  initialize variables, read command line and check it for errors, respond to -help, etc         */
   /* =============================================================================================== */
   G. = ''
   parse source . . G.0me
   G.0me = filespec('name', G.0me)
   G.0meID = '['G.0me']'
   G.0BlankMe = copies(' ', length(G.0me))
   G.0BlankMeID = copies(' ', length(G.0meID))
   G.0NameLen = 17

   /*    the 'OutText' array will hold the text to be displayed                                       */
   OutText. = ''
   OutText.0 = 0
   /*    (Shared) are variables that should be available to any subroutine that might need them       */
   Shared = 'G. Opt. OutText. Str. rpmPkgN rpmPkgNShort rpmOpts rpmErrorFile' ,
            'sigl Logfile OutFile DefaultSetList TargetSwitch'

   Opt. = ''
   Opt.0progress = 1
   ValidSwitches = 'progress taglist sets tags debug rpmdebug logfile outfile showargs test'
   interpret GetOpts(Mainargs, ValidSwitches)
   if Opt.0Errors > 0 then call Egress 'usage'

   if Opt.0showargs \= '' then do
         call ShowGetOpts
         call Egress 'okay'
      end

   if Opt.0help \= '' then do
         call Help
         call Egress 'usage'
      end
   if ArgC > 1 then do
         say G.0meID 'Error: more than one filespec; I cannot do that yet. The command line was:'
         say G.0BlankMeID'     "'Mainargs'"'
         say ''
         call Egress 'usage'
      end

   /*  change comma-separated lists to space-separated                                                */
   Opt.0sets = translate(Opt.0sets, ' ', ',')
   Opt.0tags = translate(Opt.0tags, ' ', ',')

   if Opt.0logfile \= '' then
      Logfile = Opt.0logfile

   if Opt.0outfile \= '' then
      Outfile = Opt.0outfile

   if ArgC < 1 then do
         say G.0meID' Error: missing file name'
         call Help
         call Egress 'usage'
      end

   rpmOpts = '--nosignature'
   rpmErrorFile = ''
   if Opt.0rpmdebug \= '' then do
         Opt.0debug = 1
         rpmOpts = rpmOpts||' -vv'
         rpmErrorFile = Logfile||'_rpm.err'
         call LogIt 'debugging messages from rpm.exe will go to' rpmErrorFile
      end

   /*  if no command-line options are given, show tags from the default Set list.                     */
   /*  Otherwise, give the users just what they ask for. They can still use                           */
   /*           -taglist tags=<list> sets=<list>                                                      */
   /*  if that is really what they want                                                               */
   if Opt.0taglist \= '' then DefaultSetList = ''
   if Opt.0tags \= '' then DefaultSetList = ''

   /*    get list of filename(s) to read                                                              */
   inFspec = translate(ArgV.1, "\", "/")
   /*     ret = SysFileTree(inFspec, 'Flist', 'FO', '*****')                                          */
   ret = SysFileTree(inFspec, 'Flist', 'BO', '*****')
   if ret \= 0 then do
         call LogIt ''
         call LogIt 'ERROR: SysFileTree reported error code' ret
         call Egress 'usage'
      end

   if Flist.0 == 0 then do
         call LogIt 'No files match "'inFspec'"'
         call Egress 'usage'
      end
   return 0
   /* === End of Init =============================================================================== */

   PrintArray: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    prints the OutText. array to screen                                                          */
   /*    or rather, the array whose name is passed as a parameter                                     */
   /* =============================================================================================== */
   parse arg StemN
   ArrayN = value('stemN')
   MaxC = value(ArrayN||0)

   do c = 1 to MaxC
      avarN = ArrayN||c
      call lineout Outfile, value(avarN)
   end
   call lineout Outfile
   return
   /* === End of PrintArray ========================================================================= */

   AddTagList: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    simply makes a list of all tags in the file                                                  */
   /* =============================================================================================== */

   /*    Fill out the form.  The 'add.' stem contains all of the arguments for rpm.exe.               */
   /*    This a variable that is local to this procedure, but it will be inherited by rpmToArray.     */
   if Opt.0debug > 0 then call LogIt 'AddTagList called from line 'sigl

   add. = ''
   add.0querySwitch     = '--querytags'
   add.0pkgSwitch       = TargetSwitch
   add.0qfSwitch        = ''
   add.0querySelection  = ''
   add.0qfString        = ''
   /* =============================================================================================== */
   if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddTagList'
   ret = rpmToArray()
   if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
   call Out Str.0divSet
   return
   /* === End of AddTagList ========================================================================= */

   AddSets: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    QUERY SETS:                                                                                  */
   /*    rpm can query many tags. Instead of dealing with them individually I have                    */
   /*    combined them into related sets. So, for instance, the function QfPackage()                  */
   /*    returns a query string for several package-related tags.                                     */
   /*                                                                                                 */
   /*    running rpm is what takes most of the time, so I do not want to run it                       */
   /*    separately for each tag.  But doing all of them at once is likely to need                    */
   /*    a query string longer than 4OS2's command-line length limit of 1023 bytes.                   */
   /*    So I add query strings to the command line until I reach the limit, run rpm                  */
   /*    and show the partial results, and build another command line for the remaining tags.         */
   /* =============================================================================================== */

   /* ----------------------------------------------------------------------------------------------- */
   /*    add.0qfString         == formatted query string to pass to rpm                               */
   /*    SetList               == list of sets to include                                             */
   /*    StartAddL             == where in array that AddSets starts                                  */
   /*    EndAddL               == where in array that AddSets ends                                    */
   /* ----------------------------------------------------------------------------------------------- */

   if Opt.0debug > 0 then call LogIt 'AddSets called from line 'sigl
   /*    fill out the form.  The 'add.' stem contains all of the arguments that rpm.exe will get      */
   /*    Understand that this a variable that is local to this procedure,                             */
   /*    but it will be inherited by rpmToArray                                                       */
   add. = ''
   add.0querySwitch     = '--query'
   add.0pkgSwitch       = TargetSwitch
   add.0qfSwitch        = '--queryformat'
   add.0querySelection  = ''
   add.0qfString = ''

   MaxLineLength = 850
   select
      when translate(Opt.0sets) == 'NONE' then
         SetList = ''
      when Opt.0sets == '' then
         SetList = DefaultSetList
      when translate(Opt.0sets) == 'ALL' then
         SetList = DefaultSetList
      otherwise
         SetList = Opt.0sets
   end
   if Opt.0debug > 0 then call LogIt 'SetList = 'SetList

   if SetList == '' then return 0

   /*    not sure I will need startAddL                                                               */
   startAddL = OutText.0 +1

   /*  fix: this does not verify that label 'Qf'subname exists                                        */
   /*        if it does not exist, Rexx will stop with error 43 Routine not found                     */
   /*        do                                                                                       */
   /*           grep -i --color -E "qf[A-Za-z]+:" rpminfo.cmd                                         */
   /*        to find the subroutines that return query format strings                                 */
   /*        I suppose I could include a procedure that would do that at run-time...                  */
   /*        or just keep an up-to-date list                                                          */
   /*        or just let the user deal with the Rexx error                                            */

   do while SetList \= ''
      parse var SetList subname SetList
      interpret 'QStringTemp = Qf'subname'()'
      /*    if under length limit, add to format string,                                              */
      /*    otherwise run rpm.exe with a partial format string and continue                           */
      if length(add.0qfString) + length(QStringTemp) < MaxLineLength then do
            add.0qfString = add.0qfString||QStringTemp
         end
      else do
            /*  run rpm.exe for this part of string; then start the new string                        */
            if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddSets'
            ret = rpmToArray()
            if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
            if ret \== 0 then do
                  l = OutText.0
                  if pos('is not owned', OutText.l) > 0 then
                     /*  the error message will show up in the output array                           */
                     nop
                  else
                     call LogIt 'rpm.exe returned' ret', aborting this file'
                  return 1
               end
            add.0qfString = QStringTemp
         end   /*  end of one set  */
   end   /*  while SetList \= ''  */
   /*  run rpm.exe for the remainder of string                                                        */
   if add.0qfString \= '' then do
         if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddSets'
         ret = rpmToArray()
         if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
      end
   /*    end of all sets                                                                              */

   if OutText.0 > startAddL then
      call PostProcessor
   return 0
   /* === End of AddSets ============================================================================ */

   AddUserTags: /*fold00*/
   /* =============================================================================================== */
   /*    get whatever tags the user asked for                                                         */
   /*    the display may be ugly, because we do not know what the values mean                         */
   /* =============================================================================================== */
   if Opt.0debug > 0 then call LogIt 'AddUserTags called from line 'sigl

   /* =============================================================================================== */
   /*    Fill out the form.  The 'add.' stem contains all of the arguments for rpm.exe.               */
   /*    This a variable that is local to this procedure, but it will be inherited by rpmToArray.     */
   add. = ''
   add.0querySwitch     = '--query'
   add.0pkgSwitch       = TargetSwitch
   add.0qfSwitch        = '--queryformat'
   add.0querySelection  = ''
   add.0qfString        = ''
   /* =============================================================================================== */

   /*    if "-tags=all", we first must ask rpm.exe for a list of all tags in this package             */
   if translate(Opt.0tags) == 'ALL' then do
         /*    flush queue                                                                            */
         do while queued() <> 0
            parse pull
         end

         if Opt.0progress > 0 then call ShowProgress G.0meID 'calling "rpm.exe --querytags" for' rpmPkgNShort

         if Opt.0test > 0 then do
               if Opt.0progress > 0 then call ShowProgress ''
               call LogIt '[Test] command is:'
               call LogIt '       "rpm.exe 'rpmOpts' --querytags' TargetSwitch rpmPkgN'"', ''
               return 0
            end

         if rpmErrorFile \= '' then
            'rpm.exe 'rpmOpts '--querytags' TargetSwitch rpmPkgN' 2>>'rpmErrorFile '| rxqueue.exe'
         else
            'rpm.exe 'rpmOpts '--querytags' TargetSwitch rpmPkgN' | rxqueue.exe'

         if Opt.0progress > 0 then call ShowProgress 'reading the answer from queue'
         Opt.0tags = ''
         do while queued() > 0
            aTag = linein('queue:')
            Opt.0tags = Opt.0tags' 'aTag
         end
         Opt.0tags = strip(Opt.0tags)
         /*    Opt.0tags now has a list of tags to be shown                                           */
      end   /*  end if "all"  */

   MaxLineLength = 900
   TagList = Opt.0tags
   do while TagList \= ''
      parse var TagList subname TagList
      QStringTemp = QfGenericTag(subname)
      if length(add.0qfString) + length(QStringTemp) < MaxLineLength then do
            add.0qfString = add.0qfString||QStringTemp
         end
      else do
            if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddUserTags'
            ret = rpmToArray()
            if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
            add.0qfString = QStringTemp
         end
   end
   if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddUserTags'
   if add.0qfString \= '' then do
         ret = rpmToArray()
         if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
      end
   return
   /* === End of AddUserTags ======================================================================== */

   AddTemplate: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    this is just a template for new Add* procedures                                              */
   /*    runs an rpm.exe query and text to the 'OutText.' array                                       */
   /* =============================================================================================== */

   /*    Fill out the form.  The 'add.' stem contains all of the arguments for rpm.exe.               */
   /*    This a variable that is local to this procedure, but it will be inherited by rpmToArray.     */
   add. = ''
   add.0querySwitch     = '--query'
   add.0pkgSwitch       = TargetSwitch
   add.0qfSwitch        = '--queryformat'
   add.0querySelection  = ''
   add.0qfString        = ''
   /* ----------------------------------------------------------------------------------------------- */

   /*    if using formatstrings-- for each subname, get a temp string and add it to formatstring.     */
   /*    If formatstring is too long, call rpm.exe to do part of it                                   */
   do while TagList \= ''
      parse var TagList subname TagList
      interpret 'QStringTemp = Qf'subname'()'
      if length(add.0qfString) + length(QStringTemp) < MaxLineLength then do
            add.0qfString = add.0qfString||QStringTemp
         end
      else do
            if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddTemplate'
            ret = rpmToArray()
            if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
            add.0qfString = QStringTemp
         end
   end
   if Opt.0debug > 0 then call LogIt 'call rpmToArray from AddTemplate'
   if add.0qfString \= '' then do
         ret = rpmToArray()
         if Opt.0debug > 0 then call LogIt 'rpmToArray returned' ret
      end
   return
   /* === End of AddTemplate ======================================================================== */

   QfPackage: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  package-related tags                                                                           */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'PACKAGE:\n'||Set.prefix
   tag. = ''

   /*  NAME                                                                                           */
   tag.ID = '{NAME}'
   tag.width = ''
   tag.prefix = left('Name:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  NEVRA                                                                                          */
   tag.ID = '{NEVRA}'
   tag.width = ''
   tag.prefix = left('Version:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  GROUP                                                                                          */
   tag.ID = '{GROUP}'
   tag.width = ''
   tag.prefix = left('Group:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  SUMMARY                                                                                        */
   tag.ID = '{SUMMARY}'
   tag.width = ''
   tag.prefix = left('Summary:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  URL                                                                                            */
   tag.ID = '{URL}'
   tag.width = ''
   tag.prefix = left('URL', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  DESCRIPTION                                                                                    */
   /*  indenting description is done by PostProcessor                                                 */
   tag.ID = '{DESCRIPTION}'
   tag.width = ''
   tag.prefix = left('Description:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  VENDOR                                                                                         */
   tag.ID = '{VENDOR}'
   tag.width = ''
   tag.prefix = left('Vendor:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  PACKAGER                                                                                       */
   tag.ID = '{PACKAGER}'
   tag.width = ''
   tag.prefix = left('Packager:', G.0NameLen)
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfPackage ========================================================================== */

   QfChangelog: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Changelog-related tags                                                                         */
   /*    entries in changelog can be of several lines, with embedded linefeeds.                       */
   /* =============================================================================================== */

   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'CHANGELOG:\n'||Set.prefix
   tag. = ''

   /*  CHANGELOGTIME                                                                                  */
   tag.ID = '{CHANGELOGTIME:day}'
   tag.width = '-16'
   tag.prefix = '* '
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  CHANGELOGNAME                                                                                  */
   tag.ID = '{CHANGELOGNAME}'
   tag.width = ''
   tag.prefix = ' '
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  CHANGELOGTEXT                                                                                  */
   tag.ID = '{CHANGELOGTEXT}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = '\n\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfChangelog ======================================================================== */

   QfTemplate: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  an empty template                                                                              */
   /* =============================================================================================== */
   Set. = ''
   tag. = ''

   Set.prefix = '['
   Set.suffix = ']'

   /*  TAG1                                                                                           */
   tag.ID = '{}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  TAG2                                                                                           */
   tag.ID = '{}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  TAG3                                                                                           */
   tag.ID = '{}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfTemplate ========================================================================= */

   QfFiles: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  files-related tags                                                                             */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = '  '||left('Date', 16)||right('Size', 10)||'  '||'Name\n'||Set.prefix
   Set.prefix = 'FILES:\n'||Set.prefix
   tag. = ''

   /*  FILEMTIMES                                                                                     */
   tag.ID = '{FILEMTIMES:day}'
   tag.width = '-16'
   tag.prefix = '  '
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*    not sure what the difference is between 'filesize' and 'longfilesize'                        */
   /*  FILESIZES                                                                                      */
   /*      tag.ID = '{FILESIZES}'                                                                     */
   /*      tag.width = '10'                                                                           */
   /*      tag.prefix = ''                                                                            */
   /*      tag.suffix = '  '                                                                          */
   /*      tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                                   */
   /*      Set.QF = Set.QF||tag.qf                                                                    */

   /*  LONGFILESIZES                                                                                  */
   tag.ID = '{LONGFILESIZES}'
   tag.width = '10'
   tag.prefix = ''
   tag.suffix = '  '
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  FILENAMES                                                                                      */
   tag.ID = '{FILENAMES}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  FILELINKTOS                                                                                    */
   tag.ID = '{FILELINKTOS}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfFiles ============================================================================ */

   QfGenericTag: procedure  expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    returns a simple format string for any tag that I do not know the meaning of                 */
   /* =============================================================================================== */
   parse arg TagStr

   tag. = ''
   /*  some tags are in the form "TAGNAME:modifier".  Only the first part should be uppercase         */
   parse var TagStr tagname ':' TagMod
   tagname = translate(tagname)
   if TagMod \= '' then tagname = tagname||':'||TagMod
   tag.ID = '{'tagname'}'
   tag.width = ''
   tag.prefix = 'Tag name: 'tagname'\nValue:    '
   tag.suffix = '\n'Str.0divTag'\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   tag.qf = '['tag.qf']'
   return tag.qf
   /* === End of QfGenericTag ======================================================================= */

   QfObsoletes: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  obsoletes-related tags                                                                         */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'OBSOLETES:\n'||Set.prefix

   tag. = ''
   /*  OBSOLETES                                                                                      */
   tag.ID = '{OBSOLETES}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfObsoletes ======================================================================== */

   QfProvides: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  provides-related tags                                                                          */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'PROVIDES:\n'||Set.prefix
   tag. = ''

   /*  PROVIDES                                                                                       */
   tag.ID = '{PROVIDES}'
   tag.width = ''
   tag.prefix = '  '
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  PROVIDEVERSION                                                                                 */
   tag.ID = '{PROVIDEVERSION}'
   tag.width = ''
   tag.prefix = ' '
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfProvides ========================================================================= */

   QfRecommends: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*       RECOMMEND tags                                                                            */
   /* =============================================================================================== */

   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.prefix = 'RECOMMENDS:\n'||Set.prefix
   Set.suffix = Set.suffix||Str.0SetSuffix

   tag. = ''
   /*  RECOMMENDS                                                                                     */
   tagname = 'RECOMMENDS'
   tag.ID = '{'tagname'}'
   tag.width = '-35'
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  RECOMMENDFLAGS                                                                                 */
   tagname = 'RECOMMENDFLAGS:deptype'
   tag.ID = '{'tagname'}'
   tag.width = '-15'
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  RECOMMENDNAME                                                                                  */
   tagname = 'RECOMMENDNAME'
   tag.ID = '{'tagname'}'
   tag.width = '-35'
   tag.prefix = '    '
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfRecommends ======================================================================= */

   QfRequires: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  requires-related tags                                                                          */
   /*  'requirename' seems to be the same as 'requires' (?)                                           */
   /*   'requirenevrs' seems to be the same but with versions                                         */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'REQUIRES:\n'||Set.prefix
   /*      Set.suffix = Set.suffix||'\n'                                                              */

   tag. = ''
   /*  REQUIRES                                                                                       */
   /*              tagName = 'REQUIRES'                                                               */
   /*              tag.ID = '{'tagName'}'                                                             */
   /*              tag.width = '-35'                                                                  */
   /*            tag.prefix = 'tag: 'tagName'\n'                                                      */
   /*              tag.suffix = '\n'                                                                  */
   /*              tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                           */
   /*              Set.QF = Set.QF||tag.qf                                                            */

   /*  REQUIREFLAGS                                                                                   */
   /*    tagName = 'REQUIREFLAGS:deptype'                                                             */
   /*    tag.ID = '{'tagName'}'                                                                       */
   /*    tag.width = '-15'                                                                            */
   /*    tag.prefix = 'tag: 'tagName'\n'                                                              */
   /*    tag.suffix = '\n'                                                                            */
   /*    tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                                     */
   /*    Set.QF = Set.QF||tag.qf                                                                      */

   /*  REQUIRENAME                                                                                    */
   /*         tagName = 'REQUIRENAME'                                                                 */
   /*         tag.ID = '{'tagName'}'                                                                  */
   /*         tag.width = '-35'                                                                       */
   /*       tag.prefix = 'tag: 'tagName'\n'                                                           */
   /*         tag.suffix = '\n'                                                                       */
   /*         tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix                                */
   /*         Set.QF = Set.QF||tag.qf                                                                 */

   /*  REQUIRENEVRS                                                                                   */
   tagname = 'REQUIRENEVRS'
   tag.ID = '{'tagname'}'
   tag.width = '-35'
   /*    tag.prefix = 'tag: 'tagName'\n'                                                              */
   tag.prefix = '  '
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfRequires ========================================================================= */

   QfRequiresbyfile: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    filerequire is an array that matches file list                                               */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'REQUIRES for individual files:\n'left('File:', 65, ' ')'Requires:\n'||Set.prefix
   tag. = ''

   /*  FILENAMES                                                                                      */
   tagname = 'FILENAMES'
   tag.ID = '{'tagname'}'
   tag.width = '-65'
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  FILEREQUIRE                                                                                    */
   tagname = 'FILEREQUIRE'
   tag.ID = '{'tagname'}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = '\n'
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf


   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfRequiresbyfile =================================================================== */

   QfScripts: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  anything I can find about package scripts                                                      */
   /*  fix: I tried using conditionals, which would avoid some "(none)" results.                      */
   /*       It almost worked but not quite. Come back to it later.                                    */
   /*       Maybe build it all as one string instead of using prefix and suffix?                      */
   /*       See /usr/lib/rpm/rpmpopt*                                                                 */
   /* =============================================================================================== */
   Set. = ''
   Set.prefix = '['
   Set.suffix = ']'
   Set.suffix = Set.suffix||Str.0SetSuffix
   Set.prefix = 'SCRIPT-RELATED TAGS:\n'||Set.prefix
   tag. = ''

   /*  these tags exist sometimes, but I haven't figured out what they are                            */
   /*     FILETRIGGERSCRIPTFLAGS FILETRIGGERSCRIPTPROG FILETRIGGERSCRIPTS TRANSFILETRIGGERSCRIPTFLAGS */
   /*     TRANSFILETRIGGERSCRIPTPROG TRANSFILETRIGGERSCRIPTS TRIGGERSCRIPTFLAGS                       */
   /*     TRIGGERSCRIPTPROG TRIGGERSCRIPTS VERIFYSCRIPT VERIFYSCRIPTFLAGS VERIFYSCRIPTPROG            */

   /*  if you change TagList, change Scripts string in PostProcessor also                             */
   TagList = 'PREIN PREINPROG POSTIN POSTINPROG POSTTRANS POSTTRANSPROG POSTUN POSTUNPROG'
   do while TagList \= ''
      parse var TagList tagname TagList
      tag.ID = '{'tagname'}'
      tag.width = ''
      tag.prefix = tagname':\n'
      tag.suffix = '\n'
      tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
      Set.qf = Set.qf||tag.qf
   end

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfScripts ========================================================================== */

   PostProcessor: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  If you want to make any changes to the OutText array produced by rpm.exe, this                 */
   /*  is the place to do it. I have included a couple of examples. They are very                     */
   /*  ad-hoc, depending on specific words or constructions.                                          */
   /*                                                                                                 */
   /*  this is only called by AddSets. If you want to do post-processing of other                     */
   /*  tags, either call this from the appropriate Add* routine or write a separate                   */
   /*  postprocessor                                                                                  */
   /* =============================================================================================== */
   Indented = 'DESCRIPTION:'
   Scripts = 'PREIN: PREINPROG: POSTIN: POSTINPROG: POSTTRANS: POSTTRANSPROG: POSTUN: POSTUNPROG:'
   l = 0
   do while l <= OutText.0
      l = l +1
      tagN = word(OutText.l, 1)
      if right(tagN, 1) == ':' then do
            select
               when wordpos(translate(tagN), Indented) > 0 then do
                     /*  align the package description to make it more readable:                      */
                     /*  If there is way to do this with a format string, I haven't discovered it     */
                     n = l
                     do while n <= OutText.0
                        n = n +1
                        if right(word(OutText.n, 1), 1) == ':' then leave
                        if OutText.n == Str.0divFile then leave
                        if OutText.n == Str.0divSet then leave
                        if OutText.n == Str.0divTag then leave
                        OutText.n = copies(' ', G.0NameLen)||OutText.n
                        l = n
                     end
                  end
               when wordpos(translate(tagN), Scripts) > 0 then do
                     /*  indent scripts a few spaces                                                  */
                     n = l
                     do while n <= OutText.0
                        n = n +1
                        if right(word(OutText.n, 1), 1) == ':' then leave
                        if OutText.n == Str.0divFile then leave
                        if OutText.n == Str.0divSet then leave
                        if OutText.n == Str.0divTag then leave
                        OutText.n = copies(' ', 4)||OutText.n
                        l = n
                     end
                  end
               otherwise nop
            end   /*  select  */
         end
   end   /*  do while l  */
   return 0
   /* === End of PostProcessor ====================================================================== */

   rpmToArray: procedure expose add. (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  runs rpm.exe and saves its output to the 'OutText.' array                                      */
   /*     In:  the 'add.' stem variable, which contains all of the arguments for rpm.exe              */
   /*              -inherited from calling 'Addxxx' procedure                                         */
   /*          the 'OutText.' array, which may already have lines in it                               */
   /*              -shared globally                                                                   */
   /*     Out: the 'OutText.' array, with new lines added                                             */
   /* =============================================================================================== */
   calledFrom = sigl

   /*  fix: build one string for rpm command line and use it for test, rpmdebug, normal cases         */
   rpmArgs = add.0querySwitch add.0querySelection add.0pkgSwitch rpmPkgN add.0qfSwitch ,
             '"'add.0qfString'"'
   rc = 0

   /* ----------------------------------------------------------------------------------------------- */
   /*  the whole format string would be hard to read, so break it up into lines                       */
   if Opt.0debug > 0 then do
         a1 = add.0querySwitch add.0querySelection add.0pkgSwitch rpmPkgN add.0qfSwitch
         call LogIt 'rpmToArray called from line 'calledFrom
         call LogIt 'rpm command is:'
         call lineout Logfile, '    rpm.exe 'a1
         if add.0qfString \= '' then do
               a2 = '"'add.0qfString'"'
               do while a2 \= ''
                  parse var a2 ww '\n' a2
                  if a2 \= '' then ww = ww'\n'
                  call lineout Logfile, '    'ww
               end
               call lineout Logfile, ''
            end
         call lineout Logfile
         drop a1 a2 ww
      end
   if Opt.0progress > 0 then
      call ShowProgress G.0meID 'querying' rpmPkgNShort 'with rpm.exe'
   /* ----------------------------------------------------------------------------------------------- */

   if Opt.0test > 0 then do
         if Opt.0progress > 0 then call ShowProgress ''
         call LogIt '[Test] command is:'
         call LogIt '       "@rpm.exe 'rpmOpts '<rpmArgs>' TargetSwitch rpmPkgN'"', ''
         return 0
      end

   if rpmErrorFile \= '' then
      '@rpm.exe 'rpmOpts rpmArgs '2>>'rpmErrorFile '| rxqueue.exe /fifo'
   else
      '@rpm.exe 'rpmOpts rpmArgs '| rxqueue.exe /fifo'
   if Opt.0progress > 0 then call ShowProgress ''
   if rc > 0 then call LogIt '[rpmToArray] call to rpm.exe returned 'rc
   if Opt.0debug > 0 then do
         call LogIt 'after running "rpm.exe 'rpmArgs'"'
         call LogIt 'rpm.exe returned' rc
      end
   if Opt.0progress > 0 then call ShowProgress 'reading the answer from queue'
   c = OutText.0
   do while queued() > 0
      c = c +1
      OutText.c = linein('queue:')
   end
   OutText.0 = c
   return rc
   /* === End of rpmToArray ========================================================================= */

   LogIt: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*    if Opt.0logfile is defined, saves some progress/debugging messages to a file,                */
   /*    otherwise to stderr                                                                          */
   /*    automatically adds timestamp and line number                                                 */
   /* =============================================================================================== */
   /*    IDstring = '['date() time()'] ['me 'line 'sigl']'                                            */
   IDstring = '['G.0me 'line 'sigl']'
   if arg(2, 'e') ,
    & arg(2) = '' then IDstring = copies(' ', length(IDstring))
   call lineout Logfile, IDstring arg(1)
   call lineout Logfile
   return
   /* === End of LogIt ============================================================================== */

   Out: procedure expose (Shared) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call Out <string>                                                                        */
   /*                                                                                                 */
   /*  adds <string> to the output array                                                              */
   /*  the array must already have been initialized                                                   */
   /* =============================================================================================== */
   parse arg args
   ln = OutText.0 +1
   OutText.ln = args
   OutText.0 = ln
   return 0
   /* === End of Out ================================================================================ */

   Help: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   call lineout 'stderr', ''
   call lineout 'stderr', G.0me 'shows information about an rpm archive or file or directory installed by rpm'
   call lineout 'stderr', 'Usage:'
   call lineout 'stderr', '   'G.0me '<filespec> [-option] [-option] ...'
   call lineout 'stderr', 'Options:'
   call lineout 'stderr', '    -!progress                 Turn off the progress display,'
   call lineout 'stderr', '                               but there is not much reason to do that'
   call lineout 'stderr', '    -outfile=<filename>        Send output to <file> instead of stdout'
   call lineout 'stderr', '    -taglist                   Show a list of all tags that rpm finds in the file'
   call lineout 'stderr', '                               (the same as "rpm.exe --querytags")'
   call lineout 'stderr', '                               It will not tell you what the tags mean'
   call lineout 'stderr', '    -tags="<tag> [<tag>],,,"   Show the value of each tag in the list'
   call lineout 'stderr', '                               The list must be comma- or space-separated.'
   call lineout 'stderr', '                               If there are spaces, it must be quoted.'
   call lineout 'stderr', '    -sets="<set> <set>..."     Show tags from some pre-defined sets of tags.'
   call lineout 'stderr', '                               The sets are:'
   do wctr = 1 to words(DefaultSetList)
      call lineout 'stderr', '                                     'word(DefaultSetList, wctr)
   end
   call lineout 'stderr', ''
   call lineout 'stderr', 'if no options are used, the default action is to show all sets.'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Examples:'
   call lineout 'stderr', '   'G.0me' rpm-4.13.0-20.oc00.i686.rpm'
   call lineout 'stderr', '      will tell you everything you are likely to want about the rpm package'
   call lineout 'stderr', ''
   call lineout 'stderr', '   'G.0me' F:\usr\bin\rpm.exe'
   call lineout 'stderr', '      the same, about the package that rpm.exe came from'
   call lineout 'stderr', ''
   call lineout 'stderr', '   'G.0me' F:\usr\bin\*.exe'
   call lineout 'stderr', '      the same, about every matching file'
   call lineout 'stderr', ''
   call lineout 'stderr', '   'G.0me' F:\usr\lib\rpm'
   call lineout 'stderr', '      the filename can also be a directory'
   call lineout 'stderr', ''
   call lineout 'stderr', '   'G.0me' <filespec> -sets=package,files,changelog'
   call lineout 'stderr', '      will tell you the most important things'
   call lineout 'stderr', ''
   call lineout 'stderr', '   'G.0me' <filespec> -tags=all'
   call lineout 'stderr', '      will tell you absolutely everything, in alphabetical order'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Getting information from rpm can be slow, depending on your hardware and'
   call lineout 'stderr', 'how much information you ask for.'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Options for debugging:'
   call lineout 'stderr', '    -debug'
   call lineout 'stderr', '            show a few debugging messages'
   call lineout 'stderr', '    -rpmdebug'
   call lineout 'stderr', '            add "-vv" to rpm.exe command to get more information from it'
   call lineout 'stderr', '            this implies -debug'
   call lineout 'stderr', '            stderr from rpm.exe will be saved to file <logfile>_rpm.err'
   call lineout 'stderr', '            if no logfile is given, the file will be called stderr_rpm.err'
   call lineout 'stderr', '    -logfile=<logfile>'
   call lineout 'stderr', '            send error and debug messages to <logfile> instead of stderr'
   call lineout 'stderr', '            (some messages may still go to stderr)'
   call lineout 'stderr', '    -test'
   call lineout 'stderr', '            show the rpm commands but do not execute them'
   call lineout 'stderr', '    -showargs'
   call lineout 'stderr', '            show how command line arguments were interpreted by GetOpts()'
   return
   /* === End of Help =============================================================================== */

   NoValue: /*fold00*/
   0NoValue:
   /* =============================================================================================== */
   /*  signal on NoValue                                                                              */
   /* ----------------------------------------------------------------------------------------------- */
   /*  depends on variables: G.0meID sigl                                                             */
   /*  depends on functions: Egress()                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   signal off NoValue
   _ = sigl
   call lineout 'stderr', G.0meID'[NOVALUE Handler] at line '_' for variable "'condition('d')'"'
   call lineout 'stderr', '   ==> {'sourceline(_)'}'
   call Egress 'bug_NoValue'
   /* === End of 0NoValue =========================================================================== */

   LineNo: /*fold00*/
   0LineNo:
   /* =============================================================================================== */
   /*  returns the number of the current line--    say 'now at line 'LineNo()                         */
   /* =============================================================================================== */
   _ = sigl
   return _
   /* === End of 0LineNo ============================================================================ */

   ShowGetOpts: /*fold00*/
   /* =============================================================================================== */
   /*    show what GetOpts() returned and how it was interpreted                                      */
   /* =============================================================================================== */
   say Str.0divSet
   say G.0meID' getopts returned these arguments:'
   do i = 1 to ArgC
      say '    'ArgV.i
   end
   say Str.0divSet

   G.0me = filespec('name', ArgV.0)
   /*    show the Opt.0* variables                                                                    */
   varlist = 'Opt.0optlist 'Opt.0optlist' ArgC'
   do i = 0 to ArgC
      varlist = varlist' ArgV.'i
   end
   do while varlist \= ''
      parse var varlist thisvar varlist
      call lineout 'stderr', '  'overlay('=  'value(thisvar), thisvar, 25)
   end
   return
   /* === End of ShowGetOpts ======================================================================== */

   Mark: /*fold00*/
   0Mark:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL Mark [<list of variable names>] [, <string>]                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Shows the filename and line number it was called from, the current values of the               */
   /*  variables named, and an informational string                                                   */
   /*  Arg(1) is a list of variable names                                                             */
   /*  Arg(2) can be any descriptive text                                                             */
   /*  Examples:                                                                                      */
   /*      CALL mark 'varA varB', 'At top of the "DO x = 1 to 5" loop'                                */
   /*  If no variable names are given, it will just show a message:                                   */
   /*      CALL mark '', 'At top of the "DO x = 1 to 5" loop'                                         */
   /*                                                                                                 */
   /*  This must share all the variables of whatever routine it was called from.                      */
   /*  I cannot think of a way to do it without setting at least one variable here,                   */
   /*  so reserve the variable '_' for the use of Mark and similar routines.                          */
   /* =============================================================================================== */
   _ = sigl
   call lineout 'stderr', ''
   if arg(2, 'e') ,
    & arg(2) = '' then nop
   else call lineout 'stderr', '[Mark]['value('G.0me')':'_']' arg(2)
   do _ = 1 to 99
      if word(arg(1), _) = '' then leave _
      if symbol(word(arg(1), _)) = 'VAR' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||'= "'||value(word(arg(1), _))||'"'
      if symbol(word(arg(1), _)) = 'LIT' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not defined'
      if symbol(word(arg(1), _)) = 'BAD' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not a variable'
   end _
   drop _
   return 0
   /* === End of 0Mark ============================================================================== */

   Egress: /*fold00*/
   0Egress:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     call Egress <exit type>                                                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  All exits come through here.                                                                   */
   /*  If you need to, say, clean up temporary files, do it from here                                 */
   /*                                                                                                 */
   /*  OKAY    == normal exit                                                                         */
   /*  USAGE   == user error                                                                          */
   /*  ERROR   == something else wrong                                                                */
   /*  BUG     == bug in the program                                                                  */
   /* =============================================================================================== */
   signal off NoValue
   signal off Syntax
   parse upper arg ExitCode

   /*    handle my typing mistakes                                                                    */
   if ExitCode = 'OKAY' then exit 0
   if ExitCode = '' then exit 0
   if ExitCode = 0 then exit 0
   if abbrev(ExitCode, 'HALT') then exit 1
   if abbrev(ExitCode, 'USAGE') then exit 1
   if abbrev(ExitCode, 'ERROR') then exit 3
   if abbrev(ExitCode, 'BUG') then exit 9
   /*    in case I used some other exit string and forgot to add it here:                             */
   exit 99
   /* === End of 0Egress ============================================================================ */

   /*  procedures that were external                                                                  */ /*fold00*/
   FileExt: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = FileExt(<filename>)                                                             */
   /*                                                                                                 */
   /*    returns the extension of a file name, like 4os2 "%@ext[]                                     */
   /* =============================================================================================== */
   parse arg Fname
   ret = ''
   Fname = filespec('N', Fname)
   dot = lastpos('.', Fname)
   if dot > 0 then ret = substr(Fname, dot +1)
   return ret
   /* === End of FileExt ============================================================================ */

   GetOpts: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     INTERPRET GetOpts(args [,valid])                                                            */
   /*                                                                                                 */
   /*  where 'args' is the command line to be parsed, and                                             */
   /*  'valid' is an optional space-separated list of valid switches                                  */
   /*  'help' and '?' are always valid                                                                */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     parse arg args                                                                              */
   /*     interpret GetOpts(args, 'switchA switchB')                                                  */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Switches must start with '-'. For example:                                                     */
   /*                                                                                                 */
   /*         -Foo                (Opt.0Foo == 1)                                                     */
   /*         -Foo=1              (Opt.0Foo == 1)                                                     */
   /*         -!Foo or -Foo=0     (Opt.0Foo == 0)                                                     */
   /*         -Foo=string         (Opt.0Foo == 'string')                                              */
   /*         -Foo="string"       (Opt.0Foo == 'string')                                              */
   /*         -Foo=a,b,c          (Opt.0Foo == 'a,b,c')                                               */
   /*         -Foo="a, b, c"      (Opt.0Foo == 'a, b, c')                                             */
   /*         -Foo="a b c"        (Opt.0Foo == 'a b c')                                               */
   /*         -Foo=123            (Opt.0Foo == '123')                                                 */
   /*         -Foo= or -Foo=''    (Opt.0Foo == '')                                                    */
   /*                                                                                                 */
   /*  switches must not have spaces before or after the '='                                          */
   /*  values with spaces must be in quotes. the quotes will be stripped.                             */
   /*                                                                                                 */
   /*  a switch without a value is set to '1'                                                         */
   /*        -switchA                                                                                 */
   /*  is the same as                                                                                 */
   /*        -switchA=1                                                                               */
   /*                                                                                                 */
   /*  a switch can be negated with "!", so                                                           */
   /*        -!switchA                                                                                */
   /*  is the same as                                                                                 */
   /*        -switchA=0                                                                               */
   /*                                                                                                 */
   /*     -? is the equivalent of -help                                                               */
   /*     they will both result in:                                                                   */
   /*        Opt.0help    == 1                                                                        */
   /*                                                                                                 */
   /*  Interpreting what GetOpts returns will create several variables that the                       */
   /*  caller can use--                                                                               */
   /*     Opt.              is a stem variable that holds whatever was specified by switches          */
   /*     Opt.0optlist      a space-separated list of options that were read                          */
   /*     Opt.0Errors       number of errors; that is, when GetOpts() did not understand something    */
   /*                                                                                                 */
   /*                       elements of the Opt. compound variable all start with '0', to avoid       */
   /*                       name conflicts with simple variables                                      */
   /*                                                                                                 */
   /*     ArgV.             is a stem that holds the non-switch arguments                             */
   /*                       (ArgV.0 is the name of the .cmd script, NOT the number of items           */
   /*                       in the stem)                                                              */
   /*     ArgC              is the number of non-switch arguments                                     */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     Foo.cmd -dir=up -msg="that way" -v "From here"                                              */
   /*     will result in these variables:                                                             */
   /*        Opt.0optlist == Opt.0dir Opt.0msg                                                        */
   /*        Opt.0dir     == up                                                                       */
   /*        Opt.0msg     == that way                                                                 */
   /*        Opt.0v       == 1                                                                        */
   /*        ArgV.0       == Foo.cmd                                                                  */
   /*        ArgV.1       == From here                                                                */
   /*        ArgC         == 1                                                                        */
   /*                                                                                                 */
   /*  in the case of errors, GetOpts() will set Opt.0Errors to the number of errors                  */
   /*  found, and cause the caller to show a message about it on stderr. The caller                   */
   /*  should probably abort if Opt.0Errors is greater than 0.                                        */
   /* =============================================================================================== */
   /* Some important variables (inside GetOpts) are:                                                  */
   /*    Str.SetOpts       == commands the the caller will execute to set options                     */
   /*                          in the Opt. stem variable                                              */
   /*    Str.SetArgs       == commands the the caller will execute to put the arguments               */
   /*                          in the ArgV. and ArgC variables                                        */
   /*    Str.Errors        == commands the caller will execute to report errors                       */
   /*                          reading the command line                                               */
   /*    ReturnStr         == all of the above, combined and returned to caller                       */
   /*    flag.             == variables for GetOpt's internal use                                     */
   /*                          flag.error == number of errors trying to read command line             */
   /*    Str.SwitchList    == list of options (switches) used on command line                         */
   /*    ArgC              == number of non-switch arguments                                          */
   /*    ArgV.             == array of non-switch arguments                                           */
   /*    valid             == list of valid switches                                                  */
   /*                                                                                                 */
   /* =============================================================================================== */
   flag. = ''
   flag.errors = 0
   Str. = ''
   ReturnStr = ''
   ArgC = 0
   ArgV. = ''
   parse arg args , valid
   do i = 1 to WordsQ(args)
      thisArg = WordQ(args, i)
      if left(thisArg, 1) == '-' then
         call ReadSwitch thisArg
      else do
            /*    not a switch, add to Arg list                                                       */
            ArgC = ArgC +1
            ArgV.ArgC = thisArg
         end
   end
   /* ----------------------------------------------------------------------------------------------- */
   /*    now put it all together into a series of commands for the caller to interpret                */
   if flag.errors > 0 then do
         Str.SwitchList = strip(Str.SwitchList' Opt.0Errors')
      end
   Str.SetOpts = Str.SetOpts' Opt.0Errors = 'flag.errors ';'
   Str.SetOpts = Str.SetOpts 'Opt.0optlist = "'Str.SwitchList'" ;'

   Str.SetArgs = 'ArgV. = "" ; parse source . . ArgV.0 ;  ArgC = "'ArgC'" ;'
   do i = 1 to ArgC
      varVal = Unquoted(ArgV.i)
      Str.SetArgs = Str.SetArgs' ArgV.'i' = 'Quoted(varVal)' ;'
   end

   ReturnStr = Str.errors Str.SetOpts Str.SetArgs
   return ReturnStr
   /* === End of GetOpts ============================================================================ */

   ReadSwitch: procedure expose flag. Str. valid /*fold01*/
   /* =============================================================================================== */
   /*  this is part of GetOpts()                                                                      */
   /* =============================================================================================== */
   parse arg thisArg
   Negate = 0
   switch = strip(thisArg, 'L', '-')
   if left(switch, 1) == '!' then do
         switch = strip(switch, 'L', '!')
         Negate = 1
      end
   parse var switch optName '=' optVal
   select
      /* -? and -help have fixed meanings and are allowed even if not included in 'valid'             */
      when translate(optName) == '?' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'HELP' then
         optName  = 'Opt.0Help'
      when optName == '' then do
            Str.errors = ,
              Str.errors||'call lineout "stderr", "[GetOpts] `-` without a switch name" ;'
            flag.errors = flag.errors +1
            return
         end
      when valid \= '' ,
         & wordpos(translate(optName), translate(valid)) == 0 then do
            Str.errors = Str.errors|| ,
                         'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
            flag.errors = flag.errors +1
            return
         end
      otherwise
         optName = 'Opt.0'optName
   end   /*  select  */

   select
      when Negate then
         if optVal == '' then
            optVal = 0
         else do
               Str.errors = Str.errors|| ,
                            'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
               flag.errors = flag.errors +1
               return
            end
      when optVal == '' ,
         & right(switch, 1) \= '=' then
         /*  "-switch=" means Opt.0switch == '', not 1                                                */
         optVal = '1'
      otherwise
         /*    try to be tolerant of unusual arguments; only strip quotes if they are balanced        */
         /*    and only one level of nested quotes                                                    */
         optVal = Unquoted(optVal)
   end   /*  select  */

   if wordpos(optName, Str.SetOpts) > 0 then do
         Str.errors = ,
           Str.errors||'call lineout "stderr", "[GetOpts] duplicate switch `'thisArg'`" ;'
         flag.errors = flag.errors +1
      end
   else do
         Str.SetOpts = Str.SetOpts' 'optName' = 'Quoted(optVal)' ;'
         Str.SwitchList = strip(Str.SwitchList' 'optName)
      end
   return
   /* === End of ReadSwitch ========================================================================= */

   ShowProgress: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        CALL ShowProgress <STRING> , <ROW> , <COL>                                               */
   /* ----------------------------------------------------------------------------------------------- */
   /*  writes STRING to the screen at ROW, COL and returns cursor to where it was                     */
   /*  STRING is padded or trimmed to screen width                                                    */
   /*  If ROW or COL are not given, defaults to current cursor position                               */
   /*  Call it from a loop to show progress. When done, call it with a null string                    */
   /*  to erase the progress display                                                                  */
   /*  returns 1 on success, 0 if bad parameter                                                       */
   /* =============================================================================================== */
   /*  to prevent scrolling, do not use the last column of the last row                               */

   parse value SysCurPos() with origRow origCol
   parse arg msg , row , col
   if row = '' then row = origRow
   if col = '' then col = origCol
   if \datatype(row, 'w') ,
    | \datatype(col, 'w') then return 0
   parse value SysTextScreenSize() with maxRow maxCol
   maxRow = maxRow -1 ; maxCol = maxCol -1

   if row = '' then row = origRow
   if col = '' then col = origCol
   if row > maxRow then row = maxRow
   if col > maxCol then col = maxCol
   if row < 0 then row = 0
   if col < 0 then col = 0

   msgLen = (maxCol - col)
   if row < maxRow then msgLen = msgLen +1
   msg = left(msg, msgLen)
   call SysCurPos row, col
   call charout 'SCREEN$', msg
   call SysCurPos origRow, origCol

   return 1
   /* === End of ShowProgress ======================================================================= */

   Quoted: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Quoted(string)                                                                  */
   /*                                                                                                 */
   /*    if possible, quotes a string with the kind of quotes it does not contain                     */
   /*        {He said "no"}   --> {'He said "no"'}                                                    */
   /*        {He said 'no'}   --> {"He said 'no'"}                                                    */
   /*    otherwise uses double quotes and lets the caller deal with the errors                        */
   /* =============================================================================================== */
   parse arg string
   select
      when pos('"', string) == 0 then
         newstring = '"'string'"'
      when pos("'", string) == 0 then
         newstring = "'"string"'"
      otherwise
         newstring = '"'string'"'
   end
   return newstring
   /* === End of Quoted ============================================================================= */

   Unquoted: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Unquoted(string)                                                                */
   /*                                                                                                 */
   /*    strips quotes from a string if and only if it is enclosed in a balanced pair                 */
   /*    of single or double quotes                                                                   */
   /* =============================================================================================== */
   parse arg string
   if length(string) < 2 then return string
   select
      when left(string, 1) == '"' ,
         & right(string, 1) == '"' then
         string = substr(string, 2, length(string) -2)
      when left(string, 1) == "'" ,
         & right(string, 1) == "'" then
         string = substr(string, 2, length(string) -2)
      otherwise
         nop
   end
   return string
   /* === End of Unquoted =========================================================================== */

   WordQ: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordQ(<string>, <word number>, <separator characters>, <quote characters>)            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a replacement for the Rexx Word() function.                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Word() function--                                                                     */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg Str , thisnum , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0
   Thisword  = ''
   w1        = 0
   w2        = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(Str) ,
          & pos(substr(Str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(Str) then numwords = numwords +1
   if numwords == thisnum then w1 = i

   do while (i <= length(Str))
      if pos(substr(Str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(Str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(Str) ,
                   & substr(Str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(Str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(Str, i, 1), chrsW) \= 0 ,
       & i <= length(Str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            /* unless the end of the word we want, then quit                                          */
            if w1 > 0 then do
                  w2 = i -1
                  leave
               end
            do while (pos(substr(Str, i, 1), chrsW) \= 0) ,
                   & (i <= length(Str))
               i = i +1
            end
            if i <= length(Str) then numwords = numwords +1
            /* if this is the start of the word we want, note position in string                      */
            if numwords == thisnum then w1 = i
         end   /*  wordsep  */
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   if i > length(Str) then i = length(Str)
   if w1 > 0 then do
         /* we know w1, if we ran out of string, w2 is the end of string                              */
         if w2 == 0 then w2 = i
         /* found start and end of word, copy it                                                      */
         wlen = w2 - w1 +1
         Thisword = substr(Str, w1, wlen)
      end
   return Thisword
   /* === End of WordQ ============================================================================== */

   WordsQ: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordsQ(string, <separator characters>, <quote characters>)                            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a drop-in replacement for the Rexx Word() function.                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Words() function--                                                                    */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg Str , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(Str) ,
          & pos(substr(Str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(Str) then numwords = numwords +1

   do while (i <= length(Str))
      if pos(substr(Str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(Str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(Str) ,
                   & substr(Str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(Str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(Str, i, 1), chrsW) \= 0 ,
       & i <= length(Str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            do while (pos(substr(Str, i, 1), chrsW) \= 0) ,
                   & (i <= length(Str))
               i = i +1
            end
            if i <= length(Str) then numwords = numwords +1
         end
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   return numwords
   /* === End of WordsQ ============================================================================= */

   DirExist: procedure /*fold01*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*      DirExist(<directory name>)                                                                 */
   /*                                                                                                 */
   /*  Examples:                                                                                      */
   /*      if DirExist('F:\os2') then ...                                                             */
   /*      if \DirExist('F:\NotThere') then ...                                                       */
   /*      if \DirExist(filespec('path', outfile)) then ...                                           */
   /*                                                                                                 */
   /*  returns 1 if a directory exists, 0 if it does not                                              */
   /*                                                                                                 */
   /*  use either \ or /                                                                              */
   /*  trailing slashes are stripped                                                                  */
   /*      'F:\OS2' or 'F:\OS2\'   both mean 'F:\OS2'                                                 */
   /*      'F:' or 'F:\'           both mean 'F:\'                                                    */
   /*                                                                                                 */
   /*  in 4os2.exe/cmd.exe,                                                                           */
   /*      'F:'                    means "the current directory on drive F:"                          */
   /*  but that would be pointless here; the current directory must exist                             */
   /*  likewise for                                                                                   */
   /*      "DirExist('.')"                                                                            */
   /*                                                                                                 */
   /*  if no directory name is provided at all, it is taken as '.'                                    */
   /*  that way a calling routine can verify the path of a filename without the                       */
   /*  extra step of checking the output of filespec() for ''.                                        */
   /*          if \DirExist(filespec('path', outfile)) then do                                        */
   /*              say 'invalid file name, no such directory'                                         */
   /*              return 1                                                                           */
   /*          end                                                                                    */
   /*                                                                                                 */
   /*  relative paths like "..\..\bin" with leading ".." are legal, but                               */
   /*  SysFileTree() can easily be confused by them                                                   */
   /*                                                                                                 */
   /*  wildcards are legal in <directory name>, but probably not useful                               */
   /* =============================================================================================== */

   /* ----------------------------------------------------------------------------------------------- */
   /*  the root directory of a drive needs special handling because SysFileTree()                     */
   /*  will not find it unless it has at least one subdirectory.                                      */
   /*                                                                                                 */
   /*  There is a logic to that--                                                                     */
   /*  to SysFileTree(), "F:\OS2" means "the directory named F:\OS2".                                 */
   /*  An ending slash, "F:\OS2\" means "all subdirectories of F:\OS2".                               */
   /*  so there is no way to refer to the root directory of a drive.                                  */
   /*  "R:\" means "all subdirectories under R:\", and                                                */
   /*  "R:" means "the current directory on R:".                                                      */
   /* ----------------------------------------------------------------------------------------------- */
   parse arg DirName .
   if DirName == '' then return 1
   if DirName == '.' then return 1
   DirName = translate(DirName, '\', '/')

   /*  if DirName = 'F:\' verify that drive F: exists. If so, it presumably has a root directory      */
   if right(DirName, 2) == ':\' then do
         DriveN = filespec('drive', DirName)
         if SysDriveInfo(DriveN) = '' then
            return 0
         else
            return 1
      end
   else do
         DirName = strip(DirName, 't', '\')
         ret = SysFileTree(DirName, 'Dirlist.', 'do')
         if Dirlist.0 > 0 then
            return 1
         else
            return 0
      end
   /* === End of DirExist =========================================================================== */


